# ---------------------------------------------------------------------------
# evaluate_habitat_cores.py (version 1.0 beta)
# Created on: Mon Nov 24 2008 04:08:09 PM
#
# Written by Brent L. Brock, Landscape Ecologist, Craighead Environmental Research Institute
#
# Calculates percent of core habitat remaining within evaluation area and generates a shapefile
# of the results
#
#This module depends on functionlib.py which contains functions required for processing this script
# 
# ---------------------------------------------------------------------------

# Import system modules
import sys, os, random, arcgisscripting
import functionlib as fl

# Create the Geoprocessor object
gp = arcgisscripting.create()

# Set the necessary product code
gp.SetProduct("ArcInfo")
gp.overwriteoutput = 1

# Check out spatial analyst license
gp.AddMessage("\tChecking availability of spatial analyst license...")  
if gp.CheckExtension("spatial") == "Available":
    gp.CheckOutExtension("spatial")         
else:
    gp.AddWarning("\tSpatial Analyst is needed to run this tool.  Processing aborted...")
    raise "LicenseError"

# Script arguments...

aExtent = gp.GetParameterAsText(0)
pointLayer = gp.GetParameterAsText(1)
roadLayer = gp.GetParameterAsText(2) 
minCoreSize = gp.GetParameterAsText(3) 
stinfDistance = gp.GetParameterAsText(4)
rdinfDistance = gp.GetParameterAsText(5)
outputUnits = gp.GetParameterAsText(6)
outShapefile = gp.GetParameterAsText(7)
s = str(outShapefile).rsplit("\\",1)
outWorkspace = s[0]
outTable = s[1]


# Create temporary workspaces
scratchWS = gp.scratchWorkspace
if scratchWS:
    tWorkspace = fl.CreateTempWorkspace(scratchWS)
    sWorkspace = fl.CreateTempWorkspace(scratchWS)
else:
    # Set output workspace
    gp.Workspace = outWorkspace
    tWorkspace = fl.CreateTempWorkspace(outWorkspace)
    sWorkspace = fl.CreateTempWorkspace(outWorkspace)
gp.ScratchWorkspace = tWorkspace

# Local variables...
area = minCoreSize.split(" ")
minCoreSize = str(fl.ConvertAreaToMeters(area[0], area[1]))
if outputUnits == "" or outputUnits == "#":
    outputUnits = area[1]
gp.AddMessage("Minimum Patch Size: " + str(minCoreSize) + " square meters...")
del area

stdistance = stinfDistance.split(" ")
stinfDistance = str(fl.ConvertDistanceToMeters(stdistance[0], stdistance[1]))
gp.AddMessage("Structure Distance: " + str(stinfDistance) + " meters...")
del stdistance

rddistance = rdinfDistance.split(" ")
rdinfDistance = str(fl.ConvertDistanceToMeters(rddistance[0],rddistance[1]))
gp.AddMessage("Road Distance: " + str(rdinfDistance) + " meters...")
del rddistance

cExtent = tWorkspace+"\\aExtent_copy.shp" #Temporary copy of analysis extent for manipulation
## nTable = outWorkspace+"\\"+outTable #Final output table
nTable = tWorkspace + os.sep + "xxnTable.dbf" #Temporary output table
i = 1 # Number of iterations to run

constraintLayer = aExtent

### If constraint layer is specified, make copy and clip to aExtent
##if constraintLayer == "#":
##    constraintLayer = aExtent

currentDirectory = sys.path[0]

numHouses = fl.GetNumHouses(pointLayer, aExtent)
if numHouses == 0:
    addError("No structures found within analysis area.  Verify that all inputs are in the same projection...")
else:
    gp.AddMessage(str(numHouses) + " structures found within analysis area")


# numHouses = fl.GetNumHouses(pointLayer, aExtent)

##Calculate Total Area and store in variable 'tArea'##
gp.AddMessage("Caculating area of analysis extent...")
# Make copy of analysis extent layer to modify
try:
    gp.Copy_management(aExtent, cExtent, "ShapeFile")
except:
    a = gp.MakeFeatureLayer_management(aExtent, "aLayer", "", "", "Input_FID Input_FID VISIBLE NONE")
    gp.CopyFeatures(a, cExtent)
    
# Add 'Area' Field and calculate Area
fl.CalcArea(cExtent)
tArea = fl.CalcTotalArea(gp.ScratchWorkspace, cExtent)
gp.AddMessage("Total Analysis Area = " + str(tArea))
## End Calculate Total Area ##

## Create New Table to store final results if it doesn't exist... ##

#If the nTable does not exist, create it.
gp.AddMessage("Creating output table " + nTable + "...")
if not gp.Exists(nTable):
    # Process: Create Table...
    fl.MakeTable(tWorkspace, "xxnTable.dbf")

# Get result 
gp.AddMessage("Begin simulation 1...")
sim = fl.RunSimulation(i,gp.ScratchWorkspace, aExtent, constraintLayer, numHouses, minCoreSize, stinfDistance, rdinfDistance, nTable, tArea, pointLayer, roadLayer, False, "", "", "")
result = sim[0]

# Calculate density in acres from the solution
gp.AddMessage("Calculating density...")
density = float(tArea)/int(numHouses)
density = int(round(fl.ConvertSquareMetersToOther(density, outputUnits)))
result = int(round(result, 1))

# Get cores and houses of last iteration result
gp.AddMessage("Copying shapefiles from last iteration...")

# Join area calculations to core output table
gp.AddMessage("\tAppending area calculations to core shapefile table...")
gp.QualifiedFieldNames = "UNQUALIFIED"
gp.AddField_management(nTable, "CORE", "SHORT", "", "", "", "", "NON_NULLABLE", "NON_REQUIRED", "")
gp.CalculateField_management (nTable, "CORE", "1", "VB", "")
gp.Dissolve_management (tWorkspace + "\\xxcoresLayer.shp", tWorkspace + "\\xxcoresLayer2.shp", "CORE") 
gp.MakeFeatureLayer_management (tWorkspace + "\\xxcoresLayer2.shp", "jLayer")
gp.AddJoin_management ("jLayer", "CORE", nTable, "CORE")
outShape2 = gp.CopyFeatures_Management("jLayer", outShapefile)
gp.SetParameterAsText(7, outShapefile)
params = gp.GetParameterInfo()
    
# Clean up temporary workspace
fl.CleanFiles(sWorkspace)
fl.CleanFiles(tWorkspace)
gp.Delete_management(tWorkspace,"")
gp.Delete_management(sWorkspace,"")

# Display normal result in messagebox
from Tkinter import *
import tkMessageBox
root = Tk()
root.withdraw()
tkMessageBox.showinfo(
    title="Housing Density", \
    message="Target: " + str(result) + "% of area conserved as core habitat\n Overall density: " + str(density) + " " + outputUnits + " per house")


    
